"""Console script for git adr."""
# flake8: noqa: B950
from typing import Callable, List

import cleo

from pyadr.const import STATUS_ACCEPTED, STATUS_REJECTED
from pyadr.git.core import GitAdrCore
from pyadr.git.exceptions import PyadrGitError, PyadrGitPreMergeChecksFailedError


class BaseGitCommand(cleo.Command):
    def __init__(self):
        super().__init__()
        self.git_adr_core = GitAdrCore()


class GitConfigCommand(BaseGitCommand):
    """
    Configure an ADR repository

    config
        {setting? : Configuration setting.}
        {value? : Configuration value.}
        {--l|list : List configuration settings.}
        {--u|unset : Unset configuration setting.}
    """

    def handle(self):
        if self.option("list"):
            self.git_adr_core.list_config()
        elif self.option("unset"):
            if not self.argument("setting"):
                self.line_error('Not enough arguments (missing: "words").')
            self.git_adr_core.unset_config_setting(self.argument("setting"))
        elif self.argument("setting"):
            if self.argument("value"):
                self.git_adr_core.configure(
                    self.argument("setting"), self.argument("value")
                )
            else:
                self.git_adr_core.print_config_setting(self.argument("setting"))


class GitInitCommand(BaseGitCommand):
    """
    Initialise a Git ADR repository

    init
        {--f|force : If set, will erase existing ADR directory.}
        {--a|adr-only-repo : ADR only repo. This will affect the prefixes of
                             commit messages.}
    """

    def handle(self):
        if self.option("adr-only-repo"):
            self.git_adr_core.config["git"]["adr-only-repo"] = "true"

        try:
            self.git_adr_core.git_init_adr_repo(force=self.option("force"))
        except PyadrGitError:
            return 1


class GitNewCommand(BaseGitCommand):
    """
    Create an new ADR, create a feature branch and stage new ADR in feature branch

    new
        {words* : Words in the title}
    """

    def handle(self):
        try:
            self.git_adr_core.git_new_adr(title=" ".join(self.argument("words")))
        except PyadrGitError:
            return 1


class GitProposeCommand(GitNewCommand):
    """
    Propose a new ADR (same as 'new' command)

    propose
        {words* : Words in the title}
    """


class GitAcceptCommand(BaseGitCommand):
    """
    Accept a proposed ADR by assigning an ID, updating filename, status and date, and stage to the current branch

    accept
        {file : ADR file.}
        {--t|toc : If set, generates and stages the table of content after the ADR's
                   update.}
        {--c|commit : If set, commits the updated ADR.}
    """

    def handle(self):
        self.git_adr_core.git_accept_or_reject(
            self.argument("file"),
            STATUS_ACCEPTED,
            self.option("toc"),
            self.option("commit"),
        )


class GitRejectCommand(BaseGitCommand):  # noqa
    """
    Reject a proposed ADR by assigning an ID, updating filename, status and date, and stage to the current branch

    reject
        {file : ADR file.}
        {--t|toc : If set, generates and stages the table of content after the ADR's
                   update.}
        {--c|commit : If set, commits the updated ADR.}
    """

    def handle(self):
        self.git_adr_core.git_accept_or_reject(
            self.argument("file"),
            STATUS_REJECTED,
            self.option("toc"),
            self.option("commit"),
        )


class GitCommitCommand(BaseGitCommand):
    """
    Commit an ADR

    commit
        {file : ADR file.}
    """

    def handle(self):
        self.git_adr_core.commit_adr(self.argument("file"))


class GitPreMergeChecksCommand(BaseGitCommand):
    """
    Perform sanity checks typically required on ADR files before merging a Pull Request

    pre-merge-checks
    """

    def handle(self):
        try:
            self.git_adr_core.git_pre_merge_checks()
        except PyadrGitPreMergeChecksFailedError:
            return 1


class GitHelperSlugCommand(BaseGitCommand):
    """
    Print the ADR's title in slug format

    slug
        {file : ADR file.}
    """

    def handle(self):
        self.git_adr_core.print_title_slug(self.argument("file"))


class GitHelperLowercaseCommand(BaseGitCommand):
    """
    Print the ADR's title in lowercase

    lowercase
        {file : ADR file.}
    """

    def handle(self):
        self.git_adr_core.print_title_lowercase(self.argument("file"))


class GitHelperSyncFilenameCommand(BaseGitCommand):
    """
    Sync the ADR's filename with its actual title

    sync-filename
        {file : ADR file.}
    """

    def handle(self):
        self.git_adr_core.sync_filename(self.argument("file"))


class GitHelperCommitMessageCommand(BaseGitCommand):
    """
    Print the commit message related to the ADR

    commit-message
        {file : ADR file.}
    """

    def handle(self):
        self.git_adr_core.print_commit_message(self.argument("file"))


class GitHelperBranchTitleCommand(BaseGitCommand):
    """
    Print the branch title related to the ADR's review request

    branch-title
        {file : ADR file.}
    """

    def handle(self):
        self.git_adr_core.print_branch_title(self.argument("file"))


class GitHelperCommand(BaseGitCommand):
    """
    Helper command generating and syncing various useful things

    helper
    """

    commands: List[BaseGitCommand] = []

    def __init__(self):
        self.commands.extend(
            [
                GitHelperSlugCommand(),
                GitHelperLowercaseCommand(),
                GitHelperSyncFilenameCommand(),
                GitHelperCommitMessageCommand(),
                GitHelperBranchTitleCommand(),
            ]
        )
        super().__init__()

    def handle(self):
        return self.call("help", self.config.name)


class GitGenerateTocCommand(BaseGitCommand):
    """
    Generate a table of content of the ADRs

    toc
    """

    def handle(self):
        self.git_adr_core.generate_toc()



"""Package constants"""
from pathlib import Path

###############################
# STATUSES
###############################

STATUS_ANY = "<any>"
STATUS_ANY_WITHOUT_ID = "<any status without any id yet>"
STATUS_ANY_WITH_ID = "<any status with an id>"

STATUS_PROPOSED = "proposed"
STATUS_ACCEPTED = "accepted"
STATUS_REJECTED = "rejected"
STATUS_DEPRECATED = "deprecated"
STATUS_SUPERSEDING = "superseding"
VALID_STATUSES = [
    STATUS_PROPOSED,
    STATUS_ACCEPTED,
    STATUS_REJECTED,
    STATUS_DEPRECATED,
    STATUS_SUPERSEDING,
]
STATUSES_WITH_ID = [
    STATUS_ANY_WITH_ID,
    STATUS_ACCEPTED,
    STATUS_REJECTED,
    STATUS_DEPRECATED,
    STATUS_SUPERSEDING,
]
STATUSES_WITHOUT_ID = [
    STATUS_ANY_WITHOUT_ID,
    STATUS_PROPOSED,
]

###############################
# REVIEW REQUESTS
###############################

REVIEW_REQUESTS = {
    STATUS_PROPOSED: "propose",
    STATUS_DEPRECATED: "deprecate",
    STATUS_SUPERSEDING: "supersede",
}
###############################
# CONFIG AND SETTINGS
###############################

DEFAULT_CONFIG_FILE_NAME = ".adr"
DEFAULT_CONFIG_FILE_PATH = Path(DEFAULT_CONFIG_FILE_NAME)
DEFAULT_ADR_PATH = Path("docs", "adr")

ADR_DEFAULT_SETTINGS = {"records-dir": str(DEFAULT_ADR_PATH)}

###############################
# CONTENT FORMAT
###############################

VALID_ADR_CONTENT_FORMAT = """>>>>>
# Title

* Status: a_status
[..]
* Date: YYYY-MM-DD
[..]
<<<<<
"""

###############################
# REGEX
###############################

ADR_ID_NOT_SET_REGEX = r"XXXX"
ADR_ID_REGEX = r"[0-9][0-9][0-9][0-9]"
ADR_TITLE_SLUG_REGEX = r"[a-z0-9-]*"

ADR_ID_NOT_SET_REGEX_WITH_SEPARATOR = ADR_ID_NOT_SET_REGEX + "-"
ADR_ID_REGEX_WITH_SEPARATOR = ADR_ID_REGEX + "-"

VALID_ADR_FILENAME_REGEX = (
    r"^("
    + ADR_ID_NOT_SET_REGEX
    + r"|"
    + ADR_ID_REGEX
    + r")-"
    + ADR_TITLE_SLUG_REGEX
    + r"\.md"
)
VALID_ADR_FILENAME_WITHOUT_ID_REGEX = (
    r"^" + ADR_ID_NOT_SET_REGEX + r"-" + ADR_TITLE_SLUG_REGEX + r"\.md"
)
VALID_ADR_FILENAME_WITH_ID_REGEX = (
    r"^" + ADR_ID_REGEX + r"-" + ADR_TITLE_SLUG_REGEX + r"\.md"
)

VALID_ADR_FILENAME_SKIP_TITLE_REGEX = (
    r"^(" + ADR_ID_NOT_SET_REGEX + r"|" + ADR_ID_REGEX + r")-.*\.md"
)
VALID_ADR_FILENAME_WITHOUT_ID_SKIP_TITLE_REGEX = (
    r"^" + ADR_ID_NOT_SET_REGEX + r"-.*\.md"
)
VALID_ADR_FILENAME_WITH_ID_SKIP_TITLE_REGEX = r"^" + ADR_ID_REGEX + r"-.*\.md"

FILENAME_REGEXES = {
    STATUS_ANY: {
        "full": VALID_ADR_FILENAME_REGEX,
        "skip_title": VALID_ADR_FILENAME_SKIP_TITLE_REGEX,
    },
    STATUS_ANY_WITHOUT_ID: {
        "full": VALID_ADR_FILENAME_WITHOUT_ID_REGEX,
        "skip_title": VALID_ADR_FILENAME_WITHOUT_ID_SKIP_TITLE_REGEX,
    },
    STATUS_ANY_WITH_ID: {
        "full": VALID_ADR_FILENAME_WITH_ID_REGEX,
        "skip_title": VALID_ADR_FILENAME_WITH_ID_SKIP_TITLE_REGEX,
    },
    STATUS_PROPOSED: {
        "full": VALID_ADR_FILENAME_WITHOUT_ID_REGEX,
        "skip_title": VALID_ADR_FILENAME_WITHOUT_ID_SKIP_TITLE_REGEX,
    },
    STATUS_ACCEPTED: {
        "full": VALID_ADR_FILENAME_WITH_ID_REGEX,
        "skip_title": VALID_ADR_FILENAME_WITH_ID_SKIP_TITLE_REGEX,
    },
    STATUS_REJECTED: {
        "full": VALID_ADR_FILENAME_WITH_ID_REGEX,
        "skip_title": VALID_ADR_FILENAME_WITH_ID_SKIP_TITLE_REGEX,
    },
    STATUS_DEPRECATED: {
        "full": VALID_ADR_FILENAME_WITH_ID_REGEX,
        "skip_title": VALID_ADR_FILENAME_WITH_ID_SKIP_TITLE_REGEX,
    },
    STATUS_SUPERSEDING: {
        "full": VALID_ADR_FILENAME_WITH_ID_REGEX,
        "skip_title": VALID_ADR_FILENAME_WITH_ID_SKIP_TITLE_REGEX,
    },
}

###############################
# REGEX ERROR MESSAGES
###############################

REGEX_ERROR_MESSAGE_PREFIX = (
    "(status to verify against: '{status}')\nADR(s)'s filename follow the format '"
)

REGEX_ERROR_MESSAGE_ADR_FILENAME = "".join(
    [
        REGEX_ERROR_MESSAGE_PREFIX,
        ADR_ID_NOT_SET_REGEX,
        "-<adr-title-in-slug-format>.md' or '",
        ADR_ID_REGEX,
        "-<adr-title-in-slug-format>.md'",
    ]
)
REGEX_ERROR_MESSAGE_ADR_FILENAME_WITHOUT_ID = "".join(
    [
        REGEX_ERROR_MESSAGE_PREFIX,
        ADR_ID_NOT_SET_REGEX,
        "-<adr-title-in-slug-format>.md'",
    ]
)
REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID = "".join(
    [REGEX_ERROR_MESSAGE_PREFIX, ADR_ID_REGEX, "-<adr-title-in-slug-format>.md'"]
)

REGEX_ERROR_MESSAGE_ADR_FILENAME_SKIP_TITLE = "".join(
    [
        REGEX_ERROR_MESSAGE_PREFIX,
        ADR_ID_NOT_SET_REGEX,
        "-*.md' or '",
        ADR_ID_REGEX,
        "-*.md'",
    ]
)
REGEX_ERROR_MESSAGE_ADR_FILENAME_WITHOUT_ID_SKIP_TITLE = "".join(
    [REGEX_ERROR_MESSAGE_PREFIX, ADR_ID_NOT_SET_REGEX, "-*.md'"]
)
REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID_SKIP_TITLE = "".join(
    [REGEX_ERROR_MESSAGE_PREFIX, ADR_ID_REGEX, "-*.md'"]
)

REGEX_ERROR_MESSAGES = {
    STATUS_ANY: {
        "full": REGEX_ERROR_MESSAGE_ADR_FILENAME.format(status=STATUS_ANY),
        "skip_title": REGEX_ERROR_MESSAGE_ADR_FILENAME_SKIP_TITLE.format(
            status=STATUS_ANY
        ),
        "id_prefix": "' or '".join(
            [ADR_ID_REGEX_WITH_SEPARATOR, ADR_ID_NOT_SET_REGEX_WITH_SEPARATOR]
        ),
    },
    STATUS_ANY_WITHOUT_ID: {
        "full": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITHOUT_ID.format(
            status=STATUS_ANY_WITHOUT_ID
        ),
        "skip_title": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITHOUT_ID_SKIP_TITLE.format(
            status=STATUS_ANY_WITHOUT_ID
        ),
        "id_prefix": ADR_ID_NOT_SET_REGEX_WITH_SEPARATOR,
    },
    STATUS_ANY_WITH_ID: {
        "full": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID.format(
            status=STATUS_ANY_WITH_ID
        ),
        "skip_title": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID_SKIP_TITLE.format(
            status=STATUS_ANY_WITH_ID
        ),
        "id_prefix": ADR_ID_REGEX_WITH_SEPARATOR,
    },
    STATUS_PROPOSED: {
        "full": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITHOUT_ID.format(
            status=STATUS_PROPOSED
        ),
        "skip_title": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITHOUT_ID_SKIP_TITLE.format(
            status=STATUS_PROPOSED
        ),
        "id_prefix": ADR_ID_NOT_SET_REGEX_WITH_SEPARATOR,
    },
    STATUS_ACCEPTED: {
        "full": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID.format(status=STATUS_ACCEPTED),
        "skip_title": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID_SKIP_TITLE.format(
            status=STATUS_ACCEPTED
        ),
        "id_prefix": ADR_ID_REGEX,
    },
    STATUS_REJECTED: {
        "full": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID.format(status=STATUS_REJECTED),
        "skip_title": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID_SKIP_TITLE.format(
            status=STATUS_REJECTED
        ),
        "id_prefix": ADR_ID_REGEX_WITH_SEPARATOR,
    },
    STATUS_DEPRECATED: {
        "full": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID.format(
            status=STATUS_DEPRECATED
        ),
        "skip_title": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID_SKIP_TITLE.format(
            status=STATUS_DEPRECATED
        ),
        "id_prefix": ADR_ID_REGEX_WITH_SEPARATOR,
    },
    STATUS_SUPERSEDING: {
        "full": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID.format(
            status=STATUS_SUPERSEDING
        ),
        "skip_title": REGEX_ERROR_MESSAGE_ADR_FILENAME_WITH_ID_SKIP_TITLE.format(
            status=STATUS_SUPERSEDING
        ),
        "id_prefix": ADR_ID_NOT_SET_REGEX_WITH_SEPARATOR,
    },
}



import logging

from cleo.config import ApplicationConfig
from loguru import logger

from pyadr.cli.io import LOGGING_VERY_VERBOSE, ClikitDebuggerHandler, ClikitHandler


class LoggingAppConfig(ApplicationConfig):
    def create_io(self, *args, **kwargs):
        # remove loguru's default handler
        logger.remove()

        # add cli's own handler
        io = super().create_io(*args, **kwargs)
        logger.add(ClikitDebuggerHandler(io), colorize=True, level=logging.DEBUG)
        logger.add(ClikitHandler(io), format="{message}", level=LOGGING_VERY_VERBOSE)
        logger.debug("Logger initialized.")

        return io



import re
import shutil
from pathlib import Path
from typing import List, Optional, Tuple

from loguru import logger
from slugify import slugify

from pyadr import assets
from pyadr.config import AdrConfig
from pyadr.const import (
    FILENAME_REGEXES,
    REGEX_ERROR_MESSAGES,
    STATUS_ACCEPTED,
    STATUS_ANY_WITH_ID,
    STATUS_PROPOSED,
    VALID_ADR_CONTENT_FORMAT,
    VALID_ADR_FILENAME_WITH_ID_REGEX,
)
from pyadr.content_utils import (
    adr_status_from_file,
    adr_title_lowercase_from_file,
    adr_title_slug_from_file,
    adr_title_status_and_date_from_file,
    build_toc_content_from_adrs_by_status,
    extract_adrs_by_status,
)
from pyadr.exceptions import (
    PyadrAdrDirectoryAlreadyExistsError,
    PyadrAdrDirectoryDoesNotExistsError,
    PyadrAdrFilenameFormatError,
    PyadrAdrFilenameIncorrectError,
    PyadrAdrFormatError,
    PyadrAdrRepoChecksFailedError,
    PyadrNoNumberedAdrError,
    PyadrNoProposedAdrError,
    PyadrSomeAdrFilenamesIncorrectError,
    PyadrSomeAdrIdsNotUniqueError,
    PyadrSomeAdrStatusesAreProposedError,
    PyadrTooManyProposedAdrError,
)
from pyadr.file_utils import calculate_next_adr_id, update_adr

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources  # type: ignore


class AdrCore(object):
    def __init__(self):
        self.config = AdrConfig()

    ###########################################
    # CONFIGURE ADR
    ###########################################
    def configure(self, setting: str, value: str) -> None:
        self.config["adr"][setting] = value
        logger.info(f"Configured '{setting}' to '{value}'.")

    def unset_config_setting(self, setting: str) -> None:
        del self.config["adr"][setting]
        logger.info(f"AdrConfig setting '{setting}' unset.")

    def list_config(self) -> None:
        raw_config = self.config.raw()
        for setting in raw_config.keys():
            logger.info(f"{setting} = {raw_config[setting]}")

    def print_config_setting(self, setting: str) -> None:
        raw_config = self.config.raw()
        logger.info(f"{setting} = {raw_config[setting]}")

    ###########################################
    # INIT ADR
    ###########################################
    def init_adr_repo(self, force: bool = False) -> List[Path]:
        self.verify_and_prepare_pre_init(force)
        self.create_adr_repo_dir()
        created_files = [
            self._init_adr_template(),
            self._init_adr_0000(),
            self._init_adr_0001(),
        ]
        logger.info(
            f"ADR repository successfully created at "
            f"'{Path(self.config['adr']['records-dir']).resolve()}/'."
        )
        return created_files

    def verify_and_prepare_pre_init(self, force: bool = False) -> None:
        adr_repo_abs_path = Path(self.config["adr"]["records-dir"]).resolve()
        if force:
            if adr_repo_abs_path.exists():
                logger.warning(
                    f"Directory '{adr_repo_abs_path}/' already exists. "
                    f"Used '--force' option => Erasing..."
                )
                shutil.rmtree(adr_repo_abs_path)
                logger.warning("... erased.")

        else:
            if adr_repo_abs_path.exists():
                logger.error(
                    f"Directory '{adr_repo_abs_path}/' already exists. "
                    "You can use '--force' option to erase."
                )
                raise PyadrAdrDirectoryAlreadyExistsError()

    def create_adr_repo_dir(self):
        adr_repo_path = Path(self.config["adr"]["records-dir"])
        logger.info(f"Creating ADR repo directory '{adr_repo_path}'.")
        adr_repo_path.mkdir(parents=True)
        logger.log("VERBOSE", "... done.")

    def _init_adr_template(self) -> Path:
        template_path = Path(self.config["adr"]["records-dir"], "template.md")

        logger.info(f"Copying MADR template to '{template_path}'...")
        with template_path.open("w") as f:
            f.write(pkg_resources.read_text(assets, "madr-template.md"))  # type: ignore

        logger.log("VERBOSE", "... done.")
        return template_path

    def _init_adr_0000(self) -> Path:
        return self._init_adr_file("0000-record-architecture-decisions.md")

    def _init_adr_0001(self) -> Path:
        return self._init_adr_file(
            "0001-use-markdown-architectural-decision-records.md"
        )

    def _init_adr_file(self, filename: str) -> Path:
        path = Path(self.config["adr"]["records-dir"], filename)

        logger.info(f"Creating ADR '{path}'...")
        with path.open("w") as f:
            f.write(pkg_resources.read_text(assets, filename))  # type: ignore
        update_adr(path, status=STATUS_ACCEPTED)

        logger.log("VERBOSE", "... done.")
        return path

    ###########################################
    # NEW ADR
    ###########################################
    def new_adr(self, title: str, pre_checks: bool = True) -> Path:
        if pre_checks:
            self.verify_adr_dir_exists()

        adr_path = Path(self.config["adr"]["records-dir"], f"XXXX-{slugify(title)}.md")

        logger.info(f"Creating ADR '{adr_path}'...")
        with adr_path.open("w") as f:
            f.write(pkg_resources.read_text(assets, "madr-template.md"))  # type: ignore
        update_adr(adr_path, title=title, status=STATUS_PROPOSED)
        logger.log("VERBOSE", "... done.")

        return adr_path

    ###########################################
    # ACCEPT / REJECT
    ###########################################
    def accept_or_reject(self, file: str, status: str, toc: bool = False) -> Path:
        processed_adr = self._apply_accept_or_reject_to_proposed_adr(Path(file), status)

        if toc:
            self.generate_toc()

        return processed_adr

    def _adr_filename_format_correct(
        self, path: Path, status: str = None, check_title_format: bool = True
    ) -> bool:
        status = self._resolve_status(path, status)
        full_or_skip_title = self._resolve_regex_type(check_title_format)

        return bool(
            re.compile(FILENAME_REGEXES[status][full_or_skip_title]).match(path.name)
        )

    def _resolve_status(self, path: Path, status: str = None) -> str:
        if status is None:
            status = adr_status_from_file(path)
        return status

    def _resolve_regex_type(self, check_title_format: bool = True) -> str:
        if check_title_format:
            return "full"
        else:
            return "skip_title"

    def _verify_adr_filename_format(
        self, path: Path, status: str = None, check_title_format: bool = True
    ) -> None:
        if not self._adr_filename_format_correct(path, status, check_title_format):
            status = self._resolve_status(path, status)
            full_or_skip_title = self._resolve_regex_type(check_title_format)

            logger.error(REGEX_ERROR_MESSAGES[status][full_or_skip_title] + ".")
            raise PyadrAdrFilenameFormatError(str(path))

    def _verify_one_and_only_one_proposed_adr_found(
        self, found_proposed_adrs: List[Path]
    ) -> None:
        if not len(found_proposed_adrs):
            logger.error(
                "Could not find a proposed ADR "
                "(should be of format 'docs/adr/XXXX-adr-title.md')."
            )
            raise PyadrNoProposedAdrError()

        elif len(found_proposed_adrs) > 1:
            logger.error(
                f"Can handle only 1 proposed ADR but found {len(found_proposed_adrs)}:"
            )
            for adr in found_proposed_adrs:
                logger.error(f"    => '{adr}'")
            raise PyadrTooManyProposedAdrError()

    def _get_next_adr_id(self) -> str:
        try:
            next_adr_id = calculate_next_adr_id(Path(self.config["adr"]["records-dir"]))
        except PyadrNoNumberedAdrError as e:
            logger.error(
                "There should be at least one initial accepted/rejected ADR "
                "(usually 'docs/adr/0000-record-architecture-decisions.md')."
            )
            raise PyadrNoNumberedAdrError(e)
        else:
            return next_adr_id

    def _apply_accept_or_reject_to_proposed_adr(
        self, proposed_adr: Path, status: str
    ) -> Path:
        next_adr_id = self._get_next_adr_id()
        processed_adr = self._sync_adr_filename(proposed_adr, next_adr_id)
        logger.info(f"Renamed ADR to: {processed_adr}")

        update_adr(processed_adr, status=status)
        logger.info(f"Changed ADR status to: {status}")

        return processed_adr

    def _sync_adr_filename(self, adr_path: Path, adr_id: str) -> Path:
        renamed_path = self._build_adr_filename(adr_path, adr_id)
        if adr_path != renamed_path:
            self._apply_filepath_update(adr_path, renamed_path)

        return renamed_path

    def _build_adr_filename(self, adr_path: Path, adr_id: str) -> Path:
        title_slug = adr_title_slug_from_file(adr_path)
        renamed_path = adr_path.with_name(
            "-".join([adr_id, title_slug]) + adr_path.suffix
        )
        return renamed_path

    def _apply_filepath_update(self, path: Path, renamed_path: Path) -> None:
        path.rename(renamed_path)

    ###########################################
    # GENERATE TOC
    ###########################################
    def generate_toc(self, pre_checks: bool = True) -> Path:
        if pre_checks:
            self.verify_adr_dir_exists()

        adr_paths = sorted(
            Path(self.config["adr"]["records-dir"]).glob("[0-9][0-9][0-9][0-9]-*")
        )

        adrs_by_status = extract_adrs_by_status(
            Path(self.config["adr"]["records-dir"]), adr_paths
        )

        toc_content = build_toc_content_from_adrs_by_status(adrs_by_status)

        toc_path = Path(self.config["adr"]["records-dir"], "index.md")
        with toc_path.open("w") as f:
            f.writelines(toc_content)

        logger.info(f"Markdown table of content generated in '{toc_path}'")

        return toc_path

    ###########################################
    # HELPER FUNCTIONS
    ###########################################
    def print_title_slug(self, file: str) -> None:
        logger.info(adr_title_slug_from_file(Path(file)))

    def print_title_lowercase(self, file: str) -> None:
        logger.info(adr_title_lowercase_from_file(Path(file)))

    def sync_filename(self, file: str) -> None:
        path = Path(file)

        self._verify_adr_filename_format(path, check_title_format=False)

        renamed_path = self._sync_adr_filename(path, path.stem.split("-", 1)[0])

        if path != renamed_path:
            logger.info(f"File renamed to '{str(renamed_path)}'.")
        else:
            logger.info("File name already up-to-date.")

    ###########################################
    # CHECK ADR REPO
    ###########################################
    def check_adr_repo(self, check_no_proposed: bool = True) -> None:
        at_least_one_check_failed = self._check_adr_repo(check_no_proposed)

        if at_least_one_check_failed:
            raise PyadrAdrRepoChecksFailedError
        else:
            logger.info("All checks passed.")

    def _check_adr_repo(self, check_no_proposed: bool = False) -> bool:
        at_least_one_check_failed = False

        adr_files = self._list_adr_files()

        try:
            self._check_adr_numbers_unique(adr_files)
        except PyadrSomeAdrIdsNotUniqueError:
            at_least_one_check_failed = True

        adrs_with_invalid_content_format = (
            self._filter_adrs_with_invalid_content_format(adr_files)  # noqa
        )
        if adrs_with_invalid_content_format:
            at_least_one_check_failed = True

        adr_files_checked_for_content_format = [
            file for file in adr_files if file not in adrs_with_invalid_content_format
        ]

        try:
            self._verify_adr_filenames(
                adr_files_checked_for_content_format,
                status=STATUS_ANY_WITH_ID,
                check_title_format=True,
            )
        except PyadrSomeAdrFilenamesIncorrectError:
            at_least_one_check_failed = True

        if check_no_proposed:
            try:
                self._check_no_adr_is_proposed(adr_files_checked_for_content_format)
            except PyadrSomeAdrStatusesAreProposedError:
                at_least_one_check_failed = True

        return at_least_one_check_failed

    def _check_adr_numbers_unique(self, adr_files: List[Path]) -> None:
        rex = re.compile(VALID_ADR_FILENAME_WITH_ID_REGEX)
        adrs_with_valid_filenames = [file for file in adr_files if rex.match(file.name)]
        unique_numbers = set(
            [file.stem.split("-", 1)[0] for file in adrs_with_valid_filenames]
        )
        adrs_aggregated_by_number = [
            [file for file in adrs_with_valid_filenames if file.stem.startswith(number)]
            for number in unique_numbers
        ]
        adrs_with_duplicate_number = list(
            filter(lambda x: len(x) > 1, adrs_aggregated_by_number)
        )
        if adrs_with_duplicate_number:
            logger.error(
                "ADRs must have a unique number, "
                "but the following files have the same number:"
            )
            for files in sorted(adrs_with_duplicate_number):
                logger.error(f"  => {[str(file) for file in sorted(files)]}.")
            raise PyadrSomeAdrIdsNotUniqueError

    def _filter_adrs_with_invalid_content_format(
        self, adr_files: List[Path]
    ) -> List[Path]:
        adr_files_with_invalid_content_format = []
        for adr in adr_files:
            try:
                adr_title_status_and_date_from_file(adr)
            except PyadrAdrFormatError:
                adr_files_with_invalid_content_format.append(adr)

        if adr_files_with_invalid_content_format:
            logger.error(
                "ADR must be of format:"
                "\n" + VALID_ADR_CONTENT_FORMAT + ""
                "but the following files where not:"
            )
            for file in sorted(adr_files_with_invalid_content_format):
                logger.error(f"  => '{str(file)}'.")
        return adr_files_with_invalid_content_format

    def _check_no_adr_is_proposed(self, adr_files: List[Path]) -> None:
        def adr_has_status(adr_path: Path, target_status: str) -> bool:
            (
                _,
                (status, _),
                _,
            ) = adr_title_status_and_date_from_file(adr_path)

            return status == target_status

        adrs_with_status_proposed = [
            file for file in adr_files if adr_has_status(file, STATUS_PROPOSED)
        ]

        if adrs_with_status_proposed:
            logger.error("ADR(s) must not have their status set to 'proposed', " "but:")
            for file in sorted(adrs_with_status_proposed):
                logger.error(f"  => '{str(file)}' has status 'proposed'.")
            raise PyadrSomeAdrStatusesAreProposedError

    def _verify_adr_filename(
        self, adr_path: Path, status: str = None, check_title_format: bool = True
    ) -> None:
        error_messages = self._verify_adr_filenames(
            [adr_path], status, check_title_format, log_and_raise=False
        )
        if error_messages:
            for message in error_messages:
                logger.error(message)
            raise PyadrAdrFilenameIncorrectError(adr_path)

    def _verify_adr_filenames(
        self,
        adr_files: List[Path],
        status: str = None,
        check_title_format: bool = True,
        log_and_raise: bool = True,
    ) -> Optional[List[str]]:
        """
        Verify a list of ADR filenames to make sur that:

        * the format of the ID portion of the filename is correct
        * the title portion of the filename is synched with the title of the ADR

        Args:
            adr_files: list of ADR files
            status: either `None` or a valid status ;
                    if `None` and the list contains only one file, then status will be
                    fetched from the file content ;
                    if `None` and the list contains more than one file, then status will
                    be set to `STATUS_ANY_WITH_ID`
            check_title_format: if True, will check the format of the title portion of
                                the ADR file
            log_and_raise: if True, will log errors found on filenames and will raise
                           an error

        Returns: an optional list of error messages

        """

        def title_slug_correct_and_title_slug(file: Path) -> Tuple[bool, str]:
            title_slug = adr_title_slug_from_file(file)

            try:
                title_in_filename = file.stem.split("-", 1)[1]
            except IndexError:
                title_in_filename = ""

            return title_in_filename != title_slug, title_slug

        if len(adr_files) == 1:
            status = self._resolve_status(adr_files[0], status)
        else:
            if status is None:
                status = STATUS_ANY_WITH_ID
        full_or_skip_title = self._resolve_regex_type(check_title_format)

        filenames_correctness_status = {}
        for file in sorted(adr_files):
            filenames_correctness_status[file] = {
                "format_not_valid": not self._adr_filename_format_correct(
                    file, status, check_title_format
                ),
                "with_incorrect_title_slug": title_slug_correct_and_title_slug(file),
            }

        error_messages = []
        for file, error_status in filenames_correctness_status.items():
            if error_status["format_not_valid"]:
                error_messages.append(
                    f"  => '{str(file)}' does not start with "
                    f"'{REGEX_ERROR_MESSAGES[status]['id_prefix']}'."
                )
            if error_status["with_incorrect_title_slug"][0]:  # type: ignore
                error_messages.append(
                    f"  => '{str(file)}' does not have the correct title slug "  # type: ignore  # noqa
                    f"('{error_status['with_incorrect_title_slug'][1]}')."
                    # type: ignore  # noqa
                )

        if error_messages:
            error_messages.sort()
            error_messages.insert(
                0, REGEX_ERROR_MESSAGES[status][full_or_skip_title] + ", but:"
            )
            if log_and_raise:
                for message in error_messages:
                    logger.error(message)
                raise PyadrSomeAdrFilenamesIncorrectError
            else:
                return error_messages
        else:
            return None

    def _list_adr_files(self) -> List[Path]:
        adr_files = [
            file
            for file in Path(self.config["adr"]["records-dir"]).glob("*.md")
            if file.name not in ["template.md", "index.md"]
        ]
        return adr_files

    ###########################################
    # SHARED FUNC
    ###########################################
    def verify_adr_dir_exists(self):
        adr_repo_path = Path(self.config["adr"]["records-dir"])

        logger.info(f"Verifying adr repo directory '{adr_repo_path}' exists... ")
        if not adr_repo_path.exists():
            logger.error(
                f"Directory '{adr_repo_path}/' does not exist. "
                "Initialise your ADR repo first."
            )
            raise PyadrAdrDirectoryDoesNotExistsError()
        logger.log("VERBOSE", "... done.")



"""All package specific exceptions"""
from pyadr.const import VALID_ADR_CONTENT_FORMAT


class PyadrError(Exception):
    """Base exception for errors raised by pyadr"""


class PyadrNoNumberedAdrError(PyadrError):
    """No numbered ADR was found"""


class PyadrNoLineWithSuffixError(PyadrError):
    """The searched line with suffix could not be found"""


class PyadrAdrDirectoryAlreadyExistsError(PyadrError):
    """ADR directory already exists"""


class PyadrAdrDirectoryDoesNotExistsError(PyadrError):
    """ADR directory does not exist"""


class PyadrNoProposedAdrError(PyadrError):
    """Could not find a proposed ADR"""


class PyadrTooManyProposedAdrError(PyadrError):
    """Too many proposed ADR found"""


class PyadrConfigSettingNotSupported(PyadrError):
    """Config setting not supported"""


class PyadrConfigFileSettingsNotSupported(PyadrError):
    """Config settings in the config file not supported"""


class PyadrAdrFormatError(PyadrError):
    """ADR file format error"""

    def __init__(self, item_name: str, source: str):
        adr_format_message = (
            "{item} not found in ADR '{source}', "
            + "where as it should be of format:\n"
            + VALID_ADR_CONTENT_FORMAT
        )
        super(PyadrAdrFormatError, self).__init__(
            adr_format_message.format(item=item_name, source=source)
        )


class PyadrAdrTitleNotFoundError(PyadrAdrFormatError):
    """ADR title was not found in ADR file"""

    def __init__(self, source: str):
        super(PyadrAdrTitleNotFoundError, self).__init__(
            item_name="Title", source=source
        )


class PyadrAdrStatusNotFoundError(PyadrAdrFormatError):
    """ADR status was not found in ADR file"""

    def __init__(self, source: str):
        super(PyadrAdrStatusNotFoundError, self).__init__(
            item_name="Status", source=source
        )


class PyadrAdrDateNotFoundError(PyadrAdrFormatError):
    """ADR date was not found in ADR file"""

    def __init__(self, source: str):
        super(PyadrAdrDateNotFoundError, self).__init__(item_name="Date", source=source)


class PyadrSomeAdrFilenamesIncorrectError(PyadrError):
    """Check on the name of each ADR files has failed"""


class PyadrAdrFilenameIncorrectError(PyadrError):
    """ADR filename is badly formatted and/or not synched with the ADR content"""


class PyadrSomeAdrStatusesAreProposedError(PyadrError):
    """Check on the fact that no ADR status is 'proposed' has failed"""


class PyadrSomeAdrIdsNotUniqueError(PyadrError):
    """Some ADRs have the same number"""


class PyadrAdrRepoChecksFailedError(PyadrError):
    """ADR repository checks have failed"""


class PyadrAdrFilenameFormatError(PyadrError):
    """ADR filename format incorrect"""


class PyadrStatusIncompatibleWithReviewRequestError(PyadrError):
    """Cannot create a review request branch with status of the given ADR"""



import re
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, TextIO, Tuple, Union

from slugify import slugify

from pyadr.exceptions import (
    PyadrAdrDateNotFoundError,
    PyadrAdrStatusNotFoundError,
    PyadrAdrTitleNotFoundError,
    PyadrNoLineWithSuffixError,
)


def update_adr_content_title(content: str, title: str) -> str:
    return update_adr_content_title_and_status(content, title=title, status=None)


def update_adr_content_status(content: str, status: str) -> str:
    return update_adr_content_title_and_status(content, title=None, status=status)


def update_adr_content_title_and_status(
    content: str, title: Union[str, None], status: Union[str, None]
) -> str:
    if not title and not status:
        raise TypeError("argument 'title' of 'status' has to be given")

    today = datetime.today().strftime("%Y-%m-%d")
    updated_content = content
    if title:
        updated_content = re.sub(
            r"^# .*$", f"# {title}", updated_content, 1, re.MULTILINE
        )
    if status:
        updated_content = re.sub(
            r"^\* Status: .*$", f"* Status: {status}", updated_content, 1, re.MULTILINE
        )
    updated_content = re.sub(
        r"^\* Date: .*$", f"* Date: {today}", updated_content, 1, re.MULTILINE
    )
    return updated_content


def adr_title_slug_from_file(path: Path) -> str:
    title, _, _ = adr_title_status_and_date_from_file(path)
    return slugify(title)


def adr_title_lowercase_from_file(path: Path) -> str:
    title, _, _ = adr_title_status_and_date_from_file(path)
    return title.lower()


def adr_status_from_file(path: Path) -> str:
    _, (status, _), _ = adr_title_status_and_date_from_file(path)
    return status


def adr_title_status_and_date_from_file(
    path: Path,
) -> Tuple[str, Tuple[str, Optional[str]], str]:
    with path.open() as f:
        try:
            title = extract_next_line_with_suffix_from_content_stream(
                f, "# ", stream_source=str(path)
            )
        except PyadrNoLineWithSuffixError:
            raise PyadrAdrTitleNotFoundError(source=str(path))

        try:
            full_status = extract_next_line_with_suffix_from_content_stream(
                f, "* Status:", stream_source=str(path)
            )
        except PyadrNoLineWithSuffixError:
            raise PyadrAdrStatusNotFoundError(source=str(path))

        status_phrase: Optional[str]
        try:
            status, status_phrase = full_status.split(" ", 1)
        except ValueError:
            status = full_status
            status_phrase = None
        else:
            status_phrase = status_phrase.strip()

        try:
            date = extract_next_line_with_suffix_from_content_stream(
                f, "* Date:", stream_source=str(path)
            )
        except PyadrNoLineWithSuffixError:
            raise PyadrAdrDateNotFoundError(source=str(path))

    return title, (status, status_phrase), date


def extract_next_line_with_suffix_from_content_stream(
    stream: TextIO, suffix: str, stream_source: str = "Not provided"
) -> str:
    line = "bootstrap string"
    while not line.startswith(suffix) and len(line) != 0:
        line = stream.readline()
    if len(line) == 0:
        raise PyadrNoLineWithSuffixError(
            f"No line found with suffix '{suffix}' "
            f"in content from source '{stream_source}'"
        )
    title = line[len(suffix) :].strip()
    return title


def build_toc_content_from_adrs_by_status(
    adrs_by_status: Dict[str, Dict[str, Any]]
) -> List[str]:
    toc_content = [
        "<!-- This file has been generated by `pyadr`. Manual changes will be "
        "erased at next generation. -->\n",
        "# Architecture Decision Records\n",
    ]
    for status in ["accepted", "rejected", "superseded", "deprecated"]:
        toc_content.append("\n")
        toc_content.append(f"## {adrs_by_status[status]['status-title']}\n")
        toc_content.append("\n")
        if adrs_by_status[status]["adrs"]:
            toc_content.extend(adrs_by_status[status]["adrs"])
        else:
            toc_content.append("* None\n")

    status = "non-standard"
    toc_content.append("\n")
    toc_content.append(f"## {adrs_by_status[status]['status-title']}\n")
    if adrs_by_status[status]["adrs-by-status"]:
        for value in adrs_by_status[status]["adrs-by-status"].values():
            toc_content.append("\n")
            toc_content.append(f"### {value['status-title']}\n")
            toc_content.append("\n")
            toc_content.extend(value["adrs"])
    else:
        toc_content.append("\n")
        toc_content.append("* None\n")
    return toc_content


def extract_adrs_by_status(
    records_path: Path, adr_paths: List[Path]
) -> Dict[str, Dict[str, Any]]:
    adrs_by_status: Dict[str, Dict[str, Any]] = {
        "accepted": {"status-title": "Accepted Records", "adrs": []},
        "rejected": {"status-title": "Rejected Records", "adrs": []},
        "superseded": {"status-title": "Superseded Records", "adrs": []},
        "deprecated": {"status-title": "Deprecated Records", "adrs": []},
        "non-standard": {
            "status-title": "Records with non-standard statuses",
            "adrs-by-status": {},
        },
    }

    for adr in adr_paths:
        (
            title,
            (status, status_phrase),
            date,
        ) = adr_title_status_and_date_from_file(adr)
        id = adr.stem.split("-")[0]

        if status_phrase:
            status_supplement = f": {status} {status_phrase}"
        else:
            status_supplement = ""

        adr_statement = (
            f"* [{id} - {title}]({adr.relative_to(records_path)}){status_supplement}\n"
        )
        try:
            adrs_by_status[status]["adrs"].append(adr_statement)
        except KeyError:
            if status not in adrs_by_status["non-standard"]["adrs-by-status"].keys():
                adrs_by_status["non-standard"]["adrs-by-status"][status] = {
                    "status-title": f"Status `{status}`",
                    "adrs": [],
                }

            adrs_by_status["non-standard"]["adrs-by-status"][status]["adrs"].append(
                adr_statement
            )
    return adrs_by_status



from copy import deepcopy

from pyadr.config import AdrConfig
from pyadr.const import ADR_DEFAULT_SETTINGS, DEFAULT_CONFIG_FILE_PATH
from pyadr.git.const import GIT_ADR_DEFAULT_SETTINGS


class GitAdrConfig(AdrConfig):
    config_file_path = DEFAULT_CONFIG_FILE_PATH

    def __init__(self):
        defaults = {**ADR_DEFAULT_SETTINGS, **GIT_ADR_DEFAULT_SETTINGS}
        sorted_defaults = {}
        for key in sorted(defaults.keys()):
            sorted_defaults[key] = defaults[key]
        super().__init__(defaults=sorted_defaults)

        self.section_defaults = {
            "adr": ADR_DEFAULT_SETTINGS,
            "git": GIT_ADR_DEFAULT_SETTINGS,
        }

        self.add_section("adr")
        self.add_section("git")

        if self.config_file_path.exists():
            self.read(self.config_file_path)

        self.check_filled_settings_supported()

    def persist(self) -> None:
        defaults = deepcopy(self.defaults())
        self[self.default_section] = {}  # type: ignore

        with self.config_file_path.open("w") as f:
            self.write(f)

        self[self.default_section] = defaults  # type: ignore



import logging

from clikit.api.io import flags as verbosity
from loguru import logger

LOGGING_VERBOSE = 18
LOGGING_VERY_VERBOSE = 16

logger.level("VERBOSE", LOGGING_VERBOSE, color="<bold>", icon="🔈️")
logger.level("VERY_VERBOSE", LOGGING_VERY_VERBOSE, color="<bold>", icon="🔊")

_levels = {
    logger.level("CRITICAL").no: verbosity.NORMAL,
    logger.level("ERROR").no: verbosity.NORMAL,
    logger.level("WARNING").no: verbosity.NORMAL,
    logger.level("SUCCESS").no: verbosity.NORMAL,
    logger.level("INFO").no: verbosity.NORMAL,
    logger.level("VERBOSE").no: verbosity.VERBOSE,
    logger.level("VERY_VERBOSE").no: verbosity.VERY_VERBOSE,
    logger.level("DEBUG").no: verbosity.DEBUG,
}


class ClikitHandler(logging.Handler):
    """Logging handler that redirects all messages to clikit io object."""

    def __init__(self, io, level=logging.NOTSET):
        super().__init__(level=level)
        self.io = io

    def emit(self, record: logging.LogRecord):
        level = _levels[record.levelno]
        if self.io.verbosity <= verbosity.VERY_VERBOSE:
            self.write_line(level, record)

    def write_line(self, level: int, record: logging.LogRecord):
        if record.levelno >= logging.WARNING:
            text = record.getMessage()
            self.io.error_line(text, flags=level)
        elif self.io.verbosity >= level:
            text = record.getMessage()
            self.io.write_line(text)


class ClikitDebuggerHandler(ClikitHandler):
    """Logging handler that redirects and format debug messages to clikit io object."""

    def emit(self, record: logging.LogRecord):
        level = _levels[record.levelno]
        if self.io.verbosity > verbosity.VERY_VERBOSE:
            self.write_line(level, record)



from pathlib import Path

from pyadr.content_utils import update_adr_content_title_and_status
from pyadr.exceptions import PyadrNoNumberedAdrError


def update_adr(file: Path, title: str = None, status: str = None) -> None:
    with file.open() as f:
        updated_content = update_adr_content_title_and_status(
            f.read(), title=title, status=status
        )
    with file.open("w") as f:
        f.write(updated_content)


def calculate_next_adr_id(adr_path: Path) -> str:
    numbered_adrs = sorted(adr_path.glob("[0-9][0-9][0-9][0-9]-*"))
    if not len(numbered_adrs):
        raise PyadrNoNumberedAdrError()
    most_recent_adr = numbered_adrs.pop()
    most_recent_id = most_recent_adr.stem.split("-")[0]
    next_id = f"{int(most_recent_id) + 1:04d}"
    return next_id



from pathlib import Path
from typing import List

from git import Commit, InvalidGitRepositoryError, Repo
from gitdb.exc import BadName
from loguru import logger

from pyadr.git.exceptions import (
    PyadrGitBranchAlreadyExistsError,
    PyadrGitIndexNotEmptyError,
    PyadrGitMainBranchDoesNotExistError,
    PyadrInvalidGitRepositoryError,
)


def verify_index_empty(repo: Repo) -> None:
    logger.info("Verifying Git index is empty...")
    try:
        count_staged_files = len(repo.index.diff("HEAD"))
    except BadName:
        # HEAD does not exist => the repo is empty, so must verify index is too
        count_staged_files = len(list(repo.index.iter_blobs()))

    if count_staged_files > 0:
        logger.error("... files staged in Git index. Clean before running command.")
        raise PyadrGitIndexNotEmptyError()

    logger.log("VERBOSE", "... done.")


def verify_branch_does_not_exist(repo: Repo, branch: str) -> None:
    logger.info(f"Verifying branch '{branch}' does not exist...")
    if "master" not in repo.heads or branch not in repo.heads:
        logger.log("VERBOSE", "... done.")
    else:
        logger.error(
            f"... branch '{branch}' already exists. Clean before running command."
        )
        raise PyadrGitBranchAlreadyExistsError(branch)


def verify_main_branch_exists(repo: Repo, branch: str = "master") -> None:
    logger.info(f"Verifying branch '{branch}' exists... ")
    if branch in repo.heads:
        logger.log("VERBOSE", "... done.")
    else:
        message = (
            "... branch '{branch}' does not exist. {supplement}"
            "Correct before running command."
        )
        if branch == "master":
            supplement = ""
        else:
            supplement = "Your repo is empty or it was deleted. "
        logger.error(message.format(branch=branch, supplement=supplement))
        raise PyadrGitMainBranchDoesNotExistError(branch)


def get_verified_repo_client(repo_workdir: Path) -> Repo:
    try:
        repo = Repo(repo_workdir)
    except InvalidGitRepositoryError as e:
        logger.error(
            f"No Git repository found in directory '{repo_workdir}/'. "
            f"Please initialise a Git repository before running command."
        )
        raise PyadrInvalidGitRepositoryError(e)
    return repo


def create_feature_branch_and_checkout(repo: Repo, branch_name: str) -> None:
    logger.info("Switching to 'master'...")
    repo.heads.master.checkout()
    logger.log("VERBOSE", "... done.")

    logger.info(f"Creating branch '{branch_name}' and switching to it...")
    repo.create_head(branch_name)
    repo.heads[branch_name].checkout()
    logger.log("VERBOSE", "... done.")


def files_committed_in_commit(commit: Commit) -> List:
    return list(commit.stats.files.keys())



import cleo

from pyadr import __version__
from pyadr.cli.commands import (
    AcceptCommand,
    CheckAdrRepoCommand,
    ConfigCommand,
    GenerateTocCommand,
    HelperCommand,
    InitCommand,
    NewCommand,
    ProposeCommand,
    RejectCommand,
)
from pyadr.cli.config import LoggingAppConfig


class App(cleo.Application):
    def __init__(self, config=None):
        super().__init__(
            config=config or LoggingAppConfig("ADR Process Tool", __version__)
        )

        self.add(ConfigCommand())
        self.add(InitCommand())
        self.add(NewCommand())
        self.add(ProposeCommand())
        self.add(AcceptCommand())
        self.add(RejectCommand())
        self.add(GenerateTocCommand())
        self.add(CheckAdrRepoCommand())
        self.add(HelperCommand())



"""Console script for pyadr."""
from typing import List

import cleo

from pyadr.const import STATUS_ACCEPTED, STATUS_REJECTED
from pyadr.core import AdrCore
from pyadr.exceptions import PyadrAdrRepoChecksFailedError


class BaseCommand(cleo.Command):
    def __init__(self):
        super().__init__()
        self.adr_core = AdrCore()


class ConfigCommand(BaseCommand):
    """
    Configure an ADR repository

    config
        {setting? : Configuration setting.}
        {value? : Configuration value.}
        {--l|list : List configuration settings.}
        {--u|unset : Unset configuration setting.}
    """

    def handle(self):
        if self.option("list"):
            self.adr_core.list_config()
        elif self.option("unset"):
            if not self.argument("setting"):
                self.line_error('Not enough arguments (missing: "words").')
            self.adr_core.unset_config_setting(self.argument("setting"))
        elif self.argument("setting"):
            if self.argument("value"):
                self.adr_core.configure(
                    self.argument("setting"), self.argument("value")
                )
            else:
                self.adr_core.print_config_setting(self.argument("setting"))


class InitCommand(BaseCommand):
    """
    Initialise an ADR repository

    init
        {--f|force : If set, will erase existing repository.}
    """

    def handle(self):
        self.adr_core.init_adr_repo(force=self.option("force"))


class NewCommand(BaseCommand):
    """
    Create an new ADR

    new
        {words* : Words in the title.}
    """

    def handle(self):
        self.adr_core.new_adr(title=" ".join(self.argument("words")))


class ProposeCommand(NewCommand):
    """
    Propose a new ADR (same as 'new' command)

    propose
        {words* : Words in the title}
    """


class AcceptCommand(BaseCommand):
    """
    Accept a proposed ADR by assigning an ID, updating filename, status and date

    accept
        {file : ADR file.}
        {--t|toc : If set, generates also the table of content.}
    """

    def handle(self):
        self.adr_core.accept_or_reject(
            self.argument("file"), STATUS_ACCEPTED, self.option("toc")
        )


class RejectCommand(BaseCommand):
    """
    Reject a proposed ADR by assigning an ID, updating filename, status and date

    reject
        {file : ADR file.}
        {--t|toc : If set, generates also the table of content.}
    """

    def handle(self):
        self.adr_core.accept_or_reject(
            self.argument("file"), STATUS_REJECTED, self.option("toc")
        )


class GenerateTocCommand(BaseCommand):
    """
    Generate a table of content of the ADRs

    toc
    """

    def handle(self):
        self.adr_core.generate_toc()


class CheckAdrRepoCommand(BaseCommand):
    """
    Perform sanity checks typically required on ADR files before merging a Pull Request

    check-adr-repo
        {--p|no-proposed : If set, will also check that there are no proposed ADR.}
    """

    def handle(self):
        try:
            self.adr_core.check_adr_repo(self.option("no-proposed"))
        except PyadrAdrRepoChecksFailedError:
            return 1


class HelperSlugCommand(BaseCommand):
    """
    Return the ADR's title in slug format

    slug
        {file : ADR file.}
    """

    def handle(self):
        self.adr_core.print_title_slug(self.argument("file"))


class HelperLowercaseCommand(BaseCommand):
    """
    Return the ADR's title in lowercase

    lowercase
        {file : ADR file.}
    """

    def handle(self):
        self.adr_core.print_title_lowercase(self.argument("file"))


class HelperSyncFilenameCommand(BaseCommand):
    """
    Sync the ADR's filename with its actual title

    sync-filename
        {file : ADR file.}
    """

    def handle(self):
        self.adr_core.sync_filename(self.argument("file"))


class HelperCommand(BaseCommand):
    """
    Helper command generating and syncing various useful things

    helper
    """

    commands: List[BaseCommand] = []

    def __init__(self):
        self.commands.extend(
            [
                HelperSlugCommand(),
                HelperLowercaseCommand(),
                HelperSyncFilenameCommand(),
            ]
        )
        super().__init__()

    def handle(self):
        return self.call("help", self.config.name)



from pyadr.exceptions import PyadrError


class PyadrGitError(PyadrError):
    """Base exception for errors raised by pyadr"""


class PyadrInvalidGitRepositoryError(PyadrGitError):
    """Could not find a valid repository"""


class PyadrGitIndexNotEmptyError(PyadrGitError):
    """Files are staged in the index"""


class PyadrGitBranchAlreadyExistsError(PyadrGitError):
    """Branch already exists"""


class PyadrGitMainBranchDoesNotExistError(PyadrGitError):
    """Main branch (master or other specified) does not exist"""


class PyadrGitPreMergeChecksFailedError(PyadrGitError):
    """Pre-merge checks have failed"""


class PyadrGitAdrNotStagedOrCommittedError(PyadrGitError):
    """ADR file was expected to be staged or committed"""


class PyadrGitAdrNotStagedError(PyadrGitError):
    """ADR file was expected to be staged"""


class PyadrGitAdrBadFilenameFormatOrTitleError(PyadrGitError):
    """ADR filename formot incorrect or title portion different from title in file"""



import cleo

from pyadr import __version__
from pyadr.cli.config import LoggingAppConfig
from pyadr.git.cli.commands import (
    GitAcceptCommand,
    GitCommitCommand,
    GitConfigCommand,
    GitGenerateTocCommand,
    GitHelperCommand,
    GitInitCommand,
    GitNewCommand,
    GitPreMergeChecksCommand,
    GitProposeCommand,
    GitRejectCommand,
)


class App(cleo.Application):
    def __init__(self, config=None):
        super().__init__(
            config=config or LoggingAppConfig("ADR Process Tool for Git", __version__)
        )

        self.add(GitConfigCommand())
        self.add(GitInitCommand())
        self.add(GitNewCommand())
        self.add(GitProposeCommand())
        self.add(GitAcceptCommand())
        self.add(GitRejectCommand())
        self.add(GitCommitCommand())
        self.add(GitHelperCommand())
        self.add(GitPreMergeChecksCommand())
        self.add(GitGenerateTocCommand())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyadr', 'pyadr.assets', 'pyadr.cli', 'pyadr.git', 'pyadr.git.cli']

package_data = \
{'': ['*']}

install_requires = \
['cleo>=0,<1', 'gitpython>=3.1,<4.0', 'loguru>=0,<1', 'python-slugify>=4,<5']

extras_require = \
{':python_version < "3.7"': ['importlib_resources>=4.0.0,<5.0.0'],
 'bdd': ['behave4git>=0,<1', 'PyHamcrest>=2.0,<3.0'],
 'docs': ['sphinx>=2.3,<3.0',
          'sphinx-autodoc-typehints>=1.10,<2.0',
          'sphinx-autobuild>=0,<1',
          'sphinx_rtd_theme>=0,<1',
          'm2r>=0,<1'],
 'format': ['isort>=5,<6', 'seed-isort-config>=2.2,<3.0', 'black'],
 'lint': ['flake8>=3.7,<4.0',
          'flake8-bugbear>=20,<21',
          'pydocstyle>=5.0,<6.0',
          'pylint>=2.3,<3.0',
          'yapf>=0,<1'],
 'repl': ['bpython>=0,<1'],
 'test': ['pytest>=6.1,<7.0',
          'pytest-cov>=2.8,<3.0',
          'pytest-mock>=3.2,<4.0',
          'pytest-html>=2.0,<3.0',
          'pytest-asyncio>=0,<1',
          'PyHamcrest>=2.0,<3.0'],
 'type': ['mypy>=0,<1']}

entry_points = \
{'console_scripts': ['git-adr = pyadr.git.cli:main', 'pyadr = pyadr.cli:main']}

setup_kwargs = {
    'name': 'pyadr',
    'version': '0.17.2',
    'description': 'CLI to help with an ADR process lifecycle (proposal/approval/rejection/deprecation/superseeding), which used git.',
    'long_description': None,
    'author': 'Emmanuel Sciara',
    'author_email': 'emmanuel.sciara@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/opinionated-digital-center/pyadr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""Package constants"""

GIT_ADR_DEFAULT_SETTINGS = {"adr-only-repo": "false"}

PROPOSAL_REQUEST = "propose"
DEPRECATION_REQUEST = "deprecate"
SUPERSEDING_REQUEST = "supersede"
VALID_REQUESTS = [
    PROPOSAL_REQUEST,
    DEPRECATION_REQUEST,
    SUPERSEDING_REQUEST,
]



from pathlib import Path

from git import Repo
from loguru import logger
from slugify import slugify

from pyadr.const import REVIEW_REQUESTS
from pyadr.content_utils import adr_status_from_file
from pyadr.core import AdrCore
from pyadr.exceptions import PyadrStatusIncompatibleWithReviewRequestError
from pyadr.git.config import GitAdrConfig
from pyadr.git.const import PROPOSAL_REQUEST
from pyadr.git.exceptions import (
    PyadrGitAdrNotStagedError,
    PyadrGitAdrNotStagedOrCommittedError,
    PyadrGitPreMergeChecksFailedError,
)
from pyadr.git.utils import (
    create_feature_branch_and_checkout,
    get_verified_repo_client,
    verify_branch_does_not_exist,
    verify_index_empty,
    verify_main_branch_exists,
)


class GitAdrCore(AdrCore):
    def __init__(self):
        self.config = GitAdrConfig()
        self._repo = None

    ###########################################
    # PROPERTIES
    ###########################################
    @property
    def commit_message_default_prefix(self) -> str:
        if self.config["git"].getboolean("adr-only-repo"):
            return "feat(adr):"
        else:
            return "docs(adr):"

    @property
    def repo(self) -> Repo:
        if not self._repo:
            self._repo = get_verified_repo_client(Path.cwd())
        return self._repo

    ###########################################
    # CONFIGURE ADR
    ###########################################
    def configure(self, setting: str, value: str) -> None:
        self.config.configure(setting, value)
        logger.info(f"Configured '{setting}' to '{value}'.")

    def unset_config_setting(self, setting: str) -> None:
        self.config.unset(setting)
        logger.info(f"AdrConfig setting '{setting}' unset.")

    ###########################################
    # GIT INIT ADR
    ###########################################
    def git_init_adr_repo(self, force: bool = False) -> None:
        verify_index_empty(self.repo)

        init_branch_name = "adr-init-repo"
        verify_branch_does_not_exist(self.repo, init_branch_name)

        created_files = self.init_adr_repo(force=force)

        if "master" not in self.repo.heads:
            logger.info("Git repo empty. Will commit files to 'master'.")
        else:
            create_feature_branch_and_checkout(self.repo, init_branch_name)

        self.repo.index.add([str(p) for p in created_files])

        message = f"{self.commit_message_default_prefix} initialise adr repository"
        self.repo.index.commit(message)

        logger.info(
            f"Files committed to branch '{self.repo.head.ref.name}' "
            f"with commit message '{message}'."
        )
        logger.info("ADR Git repo initialised.")

    ###########################################
    # GIT NEW ADR
    ###########################################
    def git_new_adr(self, title: str, pre_checks: bool = True) -> None:
        if pre_checks:
            verify_main_branch_exists(self.repo, branch="master")

        logger.info("Switching to 'master'...")
        self.repo.heads.master.checkout()
        logger.log("VERBOSE", "... done.")

        if pre_checks:
            self.verify_adr_dir_exists()

        verify_index_empty(self.repo)

        def build_adr_branch_name(request_type: str, title: str) -> str:
            return "-".join(["adr", request_type, slugify(title)])

        adr_branch_name = build_adr_branch_name(PROPOSAL_REQUEST, title)
        verify_branch_does_not_exist(self.repo, adr_branch_name)

        new_adr_path = self.new_adr(title, pre_checks=False)

        create_feature_branch_and_checkout(self.repo, adr_branch_name)

        logger.info(f"Staging '{new_adr_path}'...")
        self.repo.index.add([str(new_adr_path)])
        logger.log("VERBOSE", "... done.")

        logger.info("New ADR added to Git repo.")

    ###########################################
    # ACCEPT / REJECT
    ###########################################
    def git_accept_or_reject(
        self, file: str, status: str, toc: bool = False, commit: bool = False
    ) -> None:
        self._verify_adr_staged_or_committed(Path(file))

        processed_adr = self.accept_or_reject(file, status, toc)

        if commit:
            self._commit_adr(processed_adr)

    def _commit_adr(self, adr_path):
        logger.info(f"Committing ADR '{adr_path}'...")
        commit_message = self._commit_message_for_adr(adr_path)
        self.repo.index.commit(commit_message)
        logger.success(f"Committed ADR '{adr_path}' with message '{commit_message}'.")

    def _commit_message_for_adr(self, adr_path: Path) -> str:
        self._verify_adr_filename(adr_path)

        adr_status = adr_status_from_file(adr_path)
        return (
            f"{self._commit_message_prefix_for_status(adr_status)} "
            f"[{adr_status}] "
            f"{adr_path.stem}"
        )

    def _commit_message_prefix_for_status(self, status: str) -> str:
        if status == "proposed":
            return "chore(adr):"
        else:
            return self.commit_message_default_prefix

    def _verify_adr_staged_or_committed(
        self, path: Path, print_error_message: bool = True
    ) -> None:
        if not (self._file_staged(path) or self._file_committed(path)):
            if print_error_message:
                logger.error(f"ADR '{path}' should be staged or committed first.")
            raise PyadrGitAdrNotStagedOrCommittedError(path)

    def _verify_adr_staged(self, path: Path, print_error_message: bool = True) -> None:
        if not (self._file_staged(path) or self._file_committed(path)):
            if print_error_message:
                logger.error(f"ADR '{path}' should be staged first.")
            raise PyadrGitAdrNotStagedError(path)

    def _file_staged(self, path):
        return str(path) in [d.a_path for d in self.repo.index.diff("HEAD")]

    def _file_committed(self, path):
        return str(path) in list(self.repo.head.commit.stats.files.keys())

    def _apply_accept_or_reject_to_proposed_adr(
        self, proposed_adr: Path, status: str
    ) -> Path:
        processed_adr = super()._apply_accept_or_reject_to_proposed_adr(
            proposed_adr, status
        )

        self.repo.index.add([str(processed_adr)])

        return processed_adr

    def _apply_filepath_update(self, path: Path, renamed_path: Path) -> None:
        try:
            self._verify_adr_staged_or_committed(path, print_error_message=False)
        except PyadrGitAdrNotStagedOrCommittedError:
            logger.debug("not staged or committed")
            super()._apply_filepath_update(path, renamed_path)
            self.repo.index.add([str(renamed_path)])
        else:
            logger.debug("staged or committed")
            self.repo.git.mv(str(path), str(renamed_path))

    ###########################################
    # COMMIT ADR
    ###########################################
    def commit_adr(self, file: str) -> None:
        adr_path = Path(file)
        self._verify_adr_staged(adr_path)
        self._commit_adr(adr_path)

    ###########################################
    # GENERATE TOC
    ###########################################
    def generate_toc(self, pre_checks: bool = True) -> Path:
        toc_path = super().generate_toc()
        self.repo.index.add([str(toc_path)])
        return toc_path

    ###########################################
    # HELPER FUNCTIONS
    ###########################################
    def print_commit_message(self, file: str) -> None:
        logger.info(self._commit_message_for_adr(Path(file)))

    def print_branch_title(self, file: str) -> None:
        logger.info(self._branch_title_from_file(Path(file)))

    def _branch_title_from_file(self, adr_path: Path) -> str:
        self._verify_adr_filename(adr_path)

        adr_status = adr_status_from_file(adr_path)
        if adr_status not in REVIEW_REQUESTS.keys():
            logger.error(
                "Can only create review request branches for ADR statuses: "
                f"{list(REVIEW_REQUESTS.keys())}."
            )
            raise PyadrStatusIncompatibleWithReviewRequestError(
                f"ADR: '{adr_path}'; status: '{adr_status}'."
            )

        return "-".join([REVIEW_REQUESTS[adr_status], adr_path.stem.split("-", 1)[1]])

    ###########################################
    # GIT PRE MERGE CHECKS
    ###########################################
    def git_pre_merge_checks(self) -> None:
        at_least_one_check_failed = self._check_adr_repo(check_no_proposed=True)

        if at_least_one_check_failed:
            raise PyadrGitPreMergeChecksFailedError
        else:
            logger.info("All checks passed.")



from configparser import ConfigParser
from copy import deepcopy
from typing import Dict

from pyadr.const import ADR_DEFAULT_SETTINGS, DEFAULT_CONFIG_FILE_PATH
from pyadr.exceptions import (
    PyadrConfigFileSettingsNotSupported,
    PyadrConfigSettingNotSupported,
)


class AdrConfig(ConfigParser):
    config_file_path = DEFAULT_CONFIG_FILE_PATH

    def __init__(self, defaults=None):
        if defaults:
            super().__init__(defaults=defaults)
        else:
            sorted_default_settings = {}
            for key in sorted(ADR_DEFAULT_SETTINGS.keys()):
                sorted_default_settings[key] = ADR_DEFAULT_SETTINGS[key]
            super().__init__(defaults=sorted_default_settings)

            self.section_defaults = {"adr": ADR_DEFAULT_SETTINGS}

            self.add_section("adr")

            if self.config_file_path.exists():
                self.read(self.config_file_path)

            self.check_filled_settings_supported()

    def set(self, section: str, setting: str, value: str = None) -> None:
        if section != self.default_section:  # type: ignore
            self.check_section_supports_setting(section, setting)
        super().set(section, setting, value)
        if section != self.default_section:  # type: ignore
            self.persist()

    def has_option(self, section: str, setting: str) -> bool:
        self.check_section_supports_setting(section, setting)
        return super().has_option(section, setting)

    def remove_option(self, section: str, setting: str) -> bool:
        self.check_section_supports_setting(section, setting)
        existed = super().remove_option(section, setting)
        if existed and section != self.default_section:  # type: ignore
            self.persist()
        return existed

    def configure(self, setting: str, value: str) -> None:
        setting_supported = False
        for section in self._sections.keys():  # type: ignore
            if self.section_supports_setting(section, setting):
                self[section][setting] = value
                setting_supported = True
        if not setting_supported:
            raise PyadrConfigSettingNotSupported(
                f"'{self.optionxform(setting)}' not in {list(self.defaults().keys())}"
            )

    def unset(self, setting: str) -> None:
        setting_supported = False
        for section in self._sections.keys():  # type: ignore
            if self.section_supports_setting(section, setting):
                del self[section][setting]
                setting_supported = True
        if not setting_supported:
            raise PyadrConfigSettingNotSupported(
                f"'{self.optionxform(setting)}' not in {list(self.defaults().keys())}"
            )

    def raw(self) -> Dict[str, str]:
        raw_config = {}
        for setting in self.defaults().keys():
            for section in self._sections.keys():  # type: ignore
                if self.section_supports_setting(section, setting):
                    raw_config[setting] = self[section][setting]
        return raw_config

    # def print_config_setting(self, setting: str) -> None:
    #     logger.info(f"{setting} = {self.config['adr'][setting]}")

    def persist(self) -> None:
        defaults = deepcopy(self.defaults())
        self[self.default_section] = {}  # type: ignore

        if self.config_file_path.exists():
            tmp_config = ConfigParser()
            tmp_config.read(self.config_file_path)
            tmp_config["adr"] = self["adr"]
            with self.config_file_path.open("w") as f:
                tmp_config.write(f)
        else:
            with self.config_file_path.open("w") as f:
                self.write(f)

        self[self.default_section] = defaults  # type: ignore

    def check_filled_settings_supported(self) -> None:
        messages = []
        for section in self._sections.keys():  # type: ignore
            try:
                self.check_section_supports_filled_settings(section)
            except PyadrConfigFileSettingsNotSupported as e:
                messages.append(str(e))
        if messages:
            raise PyadrConfigFileSettingsNotSupported(", ".join(messages))

    def check_section_supports_filled_settings(self, section: str) -> None:
        unsupported_settings = [
            self.optionxform(setting)
            for setting in self._sections[section].keys()  # type: ignore
            if not self.section_supports_setting(section, setting)
        ]
        if unsupported_settings:
            raise PyadrConfigFileSettingsNotSupported(
                f"{unsupported_settings} not in "
                f"{list(self.section_defaults[section].keys())}"
            )

    def check_section_supports_setting(self, section: str, setting: str) -> None:
        if not self.section_supports_setting(section, setting):
            raise PyadrConfigSettingNotSupported(
                f"'{self.optionxform(setting)}' not in "
                f"{list(self.section_defaults[section].keys())}"
            )

    def section_supports_setting(self, section: str, setting: str) -> bool:
        if section in self.section_defaults.keys():
            return self.optionxform(setting) in self.section_defaults[section].keys()
        else:
            return True
